/*********************************************************************
*                   (c) SEGGER Microcontroller GmbH                  *
*                        The Embedded Experts                        *
*                           www.segger.com                           *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------

File    : IP_SpeedServer_UDP.c
Purpose : Sample program for embOS & emNet
          Waits for an UDP message sent by the UDP Speed Client and
          measures UDP send speed by packets that are able to be
          sent from application.
Notes   : For compatibility with interfaces that need to connect in
          any way this sample calls connect and disconnect routines
          that might not be needed in all cases.

          This sample can be used for Ethernet and dial-up interfaces
          and is configured to use the last registered interface as
          its main interface.

          For measuring CPU load while running the test, a delay
          between sending packets is used. For Gigabit Ethernet
          a packet is sent each ~12us which means we would need
          to queue ~83 packets for sending for one OS_Delay(1).
          The test expects the system tick to be 100us which means
          queueing ~8 packets for Gigabit Ethernet.
          Calls to OS_* functions therefore have to be multiplied
          by 10. Calls to IP_* functions do not get multiplied as
          the multiplier needs to be set in the IP_OS_*.c layer
          to avoid messing up the internal timing of the IP stack
          as it expects to work with a 1ms system tick.
*/

#include "RTOS.h"
#include "BSP.h"
#include "IP.h"

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/

#define USE_RX_TASK   0  // 0: Packets are read in ISR, 1: Packets are read in a task of its own.

//
// UDP speed client sample.
//
#define PORT          50030
#define IDLE_TIMEOUT  1000   // Timeout [ms] after which the connection will be closed if no new data is received.

//
// Task priorities.
//
enum {
   TASK_PRIO_IP_SERVER = 150
  ,TASK_PRIO_IP_TASK            // Priority should be higher than all IP application tasks.
#if USE_RX_TASK
  ,TASK_PRIO_IP_RX_TASK         // Must be the highest priority of all IP related tasks.
#endif
};

//
// Task stack sizes that might not fit for some interfaces (multiples of sizeof(int)).
//
#ifndef   APP_TASK_STACK_OVERHEAD
  #define APP_TASK_STACK_OVERHEAD     0
#endif

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static IP_HOOK_ON_STATE_CHANGE _StateChangeHook;
static int                     _IFaceId;

//
// Task stacks and Task-Control-Blocks.
//
static OS_STACKPTR int _IPStack[TASK_STACK_SIZE_IP_TASK/sizeof(int)];               // Stack of the IP_Task.
static OS_TASK         _IPTCB;                                                      // Task-Control-Block of the IP_Task.

#if USE_RX_TASK
static OS_STACKPTR int _IPRxStack[TASK_STACK_SIZE_IP_RX_TASK/sizeof(int)];          // Stack of the IP_RxTask.
static OS_TASK         _IPRxTCB;                                                    // Task-Control-Block of the IP_RxTask.
#endif

static OS_STACKPTR int _ServerStack[(1024 + APP_TASK_STACK_OVERHEAD)/sizeof(int)];  // Stack of the speed server.
static OS_TASK         _ServerTCB;                                                   // Task-Control-Block of the speed server.

/*********************************************************************
*
*       Prototypes
*
**********************************************************************
*/
#ifdef __cplusplus
extern "C" {     /* Make sure we have C-declarations in C++ programs */
#endif
void MainTask(void);
#ifdef __cplusplus
}
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _OnStateChange()
*
* Function description
*   Callback that will be notified once the state of an interface
*   changes.
*
* Parameters
*   IFaceId   : Zero-based interface index.
*   AdminState: Is this interface enabled ?
*   HWState   : Is this interface physically ready ?
*/
static void _OnStateChange(unsigned IFaceId, U8 AdminState, U8 HWState) {
  //
  // Check if this is a disconnect from the peer or a link down.
  // In this case call IP_Disconnect() to get into a known state.
  //
  if (((AdminState == IP_ADMIN_STATE_DOWN) && (HWState == 1)) ||  // Typical for dial-up connection e.g. PPP when closed from peer. Link up but app. closed.
      ((AdminState == IP_ADMIN_STATE_UP)   && (HWState == 0))) {  // Typical for any Ethernet connection e.g. PPPoE. App. opened but link down.
    IP_Disconnect(IFaceId);                                       // Disconnect the interface to a clean state.
  }
}

/*********************************************************************
*
*       _Send
*/
static void _Send(U32 IPAddr, U32 PacketSize, U32 TimeToSend, U8 Handle) {
  IP_PACKET* pPacket;
  U8*        pData;
  U32        Timeout;
  U32        SendCnt;
  U32        NumPacketsSent;
  U32        PacketSizeOverhead;
  int        r;

  SendCnt            = 0;
  NumPacketsSent     = 0;
  Timeout            = IP_OS_GET_TIME() + TimeToSend;
  PacketSizeOverhead = PacketSize + 28 + IP_GetIFaceHeaderSize(_IFaceId);  // Packet size with all headers is payload + IP/UDP(28) + transport/Ethernet header (14 bytes + 2 bytes padding).
  do {
    //
    // Avoid trying to allocate a packet if nothing is free
    // in the stack. This avoid warnings in DEBUG builds.
    // We do not use a delay here as this should not happen
    // with a sufficient buffer configuration for the
    // selected transport medium.
    //
    if (IP_GetFreePacketCnt(PacketSizeOverhead) == 0) {
      IP_OS_WaitNetEventTimed(0);
    }
    //
    // Prepare packet to send.
    //
    pPacket = IP_UDP_AllocEx(_IFaceId, PacketSize);
    if (pPacket != NULL) {
      pData    = (U8*)IP_UDP_GetDataPtr(pPacket);
      *pData++ = Handle;              // Response handle.
      *pData++ = 'S' + 1;             // Target is sending.
      IP_MEMCPY(pData, &SendCnt, 4);  // Current send count without the package that is about to be sent.
      //
      // Send packet. Retry sending if it fails until success or timeout.
      //
      do {
        r = IP_UDP_Send(_IFaceId, IPAddr, PORT, PORT, pPacket);
      } while ((IP_IsExpired(Timeout) == 0) && (r < 0));
      if (r >= 0) {
        SendCnt += PacketSize;
        NumPacketsSent++;
      } else {
        //
        // Free last packet on error.
        //
        IP_UDP_Free(pPacket);
      }
    }
    //
    // Send as many data as possible at once. Then
    // sleep for 100us for CPU load measurement.
    //
    if (IP_GetFreePacketCnt(PacketSizeOverhead) == 0) {
      OS_Delay(1);
    }
  } while (IP_IsExpired(Timeout) == 0);
  IP_OS_Delay(500);  // Give the other side a short time to recover if necessary.
  //
  // Send result.
  //
  pPacket = IP_UDP_AllocEx(_IFaceId, 14);
  if (pPacket == NULL) {
    while (1) {
      BSP_ToggleLED(0);
      OS_Delay(20);
    }
  }
  pData       = (U8*)IP_UDP_GetDataPtr(pPacket);
  *pData++    = Handle;                  // Response handle.
  *pData++    = 'S' + 2;                 // Target is sending its result.
  PacketSize += 28 + 14;                 // + IP/UDP and Ethernet headers.
  IP_MEMCPY(pData, &PacketSize, 4);      // Size of packets that have been sent incl headers.
  pData      += 4;
  IP_MEMCPY(pData, &NumPacketsSent, 4);  // Number of packets sent.
  IP_UDP_SendAndFree(_IFaceId, IPAddr, PORT, 0, pPacket);
}

/*********************************************************************
*
*       _ServerTask
*/
static void _ServerTask(void) {
  long               hSock;
  struct sockaddr_in Addr;
  U32                PacketSize;
  U32                TimeToSend;
  int                AddrLen;
  U8                 aData[6];

  PacketSize = IP_GetMaxAvailPacketSize(_IFaceId) - 28;  // Max. available payload in packets - IP/UDP headers.
  //
  // Open socket.
  //
  hSock = socket(AF_INET, SOCK_DGRAM, 0);
  if (hSock == 0) {  // Error, Could not get socket.
    while (1) {
      BSP_ToggleLED(0);
      OS_Delay(20);
    }
  } else {
    //
    // Bind socket to UDP port.
    //
    memset(&Addr, 0, sizeof(Addr));
    Addr.sin_family      = AF_INET;
    Addr.sin_port        = htons(PORT);
    Addr.sin_addr.s_addr = INADDR_ANY;
    bind(hSock, (struct sockaddr*)&Addr, sizeof(Addr));
    //
    // Wait for client to contact us.
    //
    while(1) {
      while (1) {
        AddrLen = sizeof(Addr);
        recvfrom(hSock, (char*)&aData[0], 1, 0, (struct sockaddr*)&Addr, &AddrLen);
        if (aData[0] == 'H') {  // Is this a handle sent by the UDP speed client ?
          //
          // Send our handle (client handle + 1) back.
          //
          aData[0]++;
          sendto(hSock, (const char*)&aData[0], 1, 0, (struct sockaddr*)&Addr, AddrLen);
          break;
        }
      }
      recvfrom(hSock, (char*)&aData[0], 6, 0, NULL, NULL);
      if (aData[1] == 'S') {
        IP_MEMCPY(&TimeToSend, &aData[2], 4);
        _Send(Addr.sin_addr.s_addr, PacketSize, TimeToSend, aData[0]);  // Multiply TimeToSend by 10 as we are using a 100us system tick.
      }
    }
  }
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       MainTask()
*
* Function description
*   Main task executed by the RTOS to create further resources and
*   running the main application.
*/
void MainTask(void) {
  IP_Init();
  _IFaceId = IP_INFO_GetNumInterfaces() - 1;                                           // Get the last registered interface ID as this is most likely the interface we want to use in this sample.
  OS_SetPriority(OS_GetTaskID(), TASK_PRIO_IP_TASK);                                   // For now, this task has highest prio except IP management tasks.
  OS_CREATETASK(&_IPTCB  , "IP_Task"  , IP_Task  , TASK_PRIO_IP_TASK   , _IPStack);    // Start the IP_Task.
#if USE_RX_TASK
  OS_CREATETASK(&_IPRxTCB, "IP_RxTask", IP_RxTask, TASK_PRIO_IP_RX_TASK, _IPRxStack);  // Start the IP_RxTask, optional.
#endif
  OS_CREATETASK(&_ServerTCB, "Speed server", _ServerTask, TASK_PRIO_IP_SERVER, _ServerStack);  // Start the speed server.
  IP_AddStateChangeHook(&_StateChangeHook, _OnStateChange);                            // Register hook to be notified on disconnects.
  IP_Connect(_IFaceId);                                                                // Connect the interface if necessary.
  OS_SetPriority(OS_GetTaskID(), 255);                                                 // Now this task has highest prio for real-time application. This is only allowed when this task does not use blocking IP API after this point.
  while (1) {
    BSP_ToggleLED(1);
    OS_Delay(200);
  }
}

/****** End Of File *************************************************/
