/*********************************************************************
*              SEGGER MICROCONTROLLER SYSTEME GmbH                   *
*        Solutions for real time microcontroller applications        *
**********************************************************************
*                                                                    *
*        (c) 1996-2015 SEGGER Microcontroller Systeme GmbH           *
*                                                                    *
* Internet: www.segger.com Support: support@segger.com               *
*                                                                    *
**********************************************************************
----------------------------------------------------------------------
File    : UDPSpeedClient.c
Purpose : UDP sample program to demonstrate the maximum speed that can
          be achieved using UDP transmissions. The UDP Speed Server is
          required as counterpart (typically on the target).
---------------------------END-OF-HEADER------------------------------
*/

#include <stdio.h>
#include <string.h>

#ifdef WIN32
#include <conio.h>
#include <windows.h>
#else
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <termios.h>
#endif

/*********************************************************************
*
*       Defines, configurable
*
**********************************************************************
*/

#define PORT              50030
#define TEST_DURATION     5             // Duration of the test [s].
#define RECV_BUFFER_SIZE  (256 * 1024)  // Receive buffer size. More allows the OS more time to fetch the data.

#ifdef WIN32
#define CLOSESOCKET(s)   closesocket(s)
#define SOCKLEN          int
#else
#define CLOSESOCKET(s)   close(s)
#define SOCKET           int
#define SOCKADDR_IN      struct sockaddr_in
#define SOCKLEN          socklen_t
#endif

/*********************************************************************
*
*       Static Variables
*
**********************************************************************
*/

static unsigned char _abBuffer[1472];  // Max size of a regular UDP packet payload.

#ifndef WIN32
static struct termios _OldTerm;
static struct termios _NewTerm;
#endif


/*********************************************************************
*
*       Static code
*
**********************************************************************
*/

#ifndef WIN32
/*********************************************************************
*
*       _InitTermios
*/
static void _InitTermios(void) {
  tcgetattr(0, &_OldTerm);           // grab old terminal i/o settings
  _NewTerm = _OldTerm;               // make new settings same as old settings
  _NewTerm.c_lflag &= ~ICANON;       // disable buffered i/o
  _NewTerm.c_lflag &= ~ECHO;         // disable echo mode
  tcsetattr(0, TCSANOW, &_NewTerm);  // use these new terminal i/o settings now
}

/*********************************************************************
*
*       _ResetTermios
*/
static void _ResetTermios(void) 
{
  tcsetattr(0, TCSANOW, &_OldTerm);
}

/*********************************************************************
*
*       getch
*/
static char getch(void) {
  char ch;

  _InitTermios();
  ch = getchar();
  _ResetTermios();
  return ch;
}
#endif

/*********************************************************************
*
*       _ExecIPAddressSyntaxError
*/
static void _ExecIPAddressSyntaxError(void) {
  printf("No valid IP addr. Please input in format: 192.168.0.10\n");
}

/*********************************************************************
*
*       _EatWhite
*/
static void _EatWhite(const char** ps) {
  const char* s = *ps;
  while ((*s == ' ') || (*s == '\t') || (*s == '\r') || (*s == '\n')) {
    s++;
  }
  *ps = s;
}

/*********************************************************************
*
*       _ParseDec
*/
static const char* _ParseDec(const char** ps, unsigned int* pData) {
  unsigned int Data = 0;
  int          NumDigits = 0;

  _EatWhite(ps);
  do {
    int v;
    char c;

    c = **ps;
    v =  ((c >= '0') && (c <= '9')) ? c - '0' : -1;
    if (v >= 0) {
      Data = (Data * 10) + v;
      (*ps)++;
      NumDigits++;
    } else {
      if (NumDigits == 0) {
        return "Expected a dec value";
      }
      *pData = Data;
      return NULL;
    }
  } while (1);
}

/*********************************************************************
*
*       _ParseIP
*
*  Function description
*    Returns U32 IP address out of string
*/
static unsigned int _ParseIP(const char* s, unsigned int * pAddr) {
  int           i;
  unsigned int  IPByte[4];

  _EatWhite(&s);
  if (*s) {
    for (i = 0; i <= 3; i++) {
      if (_ParseDec(&s, &IPByte[i])) {
        goto Error;
      }
      if (i < 3) {
        if (*s == '.') {
          s++;
        } else {
          goto Error;
        }
      }
    }
    *pAddr = ((IPByte[0] << 24) |
              (IPByte[1] << 16) |
              (IPByte[2] <<  8) |
               IPByte[3]);
    return 0;
  } else {
Error:
    return 1;  // Error.
  }
}

/*********************************************************************
*
*       Public code
*
**********************************************************************
*/

/*********************************************************************
*
*       main()
*/
int main(int argc, char* argv[], char* envp[]) {
#ifdef WIN32
  WORD          wVersionRequested;
  WSADATA       wsaData;
#endif
  SOCKET        hSock;
  SOCKADDR_IN   Addr;
  unsigned int  DestIP;
  unsigned int  TestDuration;
  unsigned int  PacketSize;
  unsigned int  NumPacketsSent;
  unsigned int  PacketCnt;
  SOCKLEN       AddrLen;
  int           r;
  int           Key;
  int           BufSize;
  unsigned char aData[6];
  char          ac[128];

  r            = 0;  // Assume O.K.
  hSock        = 0;
  TestDuration = TEST_DURATION * 1000;
  BufSize      = RECV_BUFFER_SIZE;
  
  printf("(c) 2007 - 2016 Segger Microcontroller GmbH & Co. KG www.segger.com\n");
  printf("UDPSpeedClient, compiled " __TIME__ " on " __DATE__ "\n");
  //
  // Get IP addr. of target.
  //
GetDestIPAddr:
  printf("\nPlease enter the addr. of the UDP speed server: ");
  fgets(ac, sizeof(ac), stdin);
  if (_ParseIP(&ac[0], &DestIP)) {
    _ExecIPAddressSyntaxError();
    goto GetDestIPAddr; // Invalid IP addr. Retry asking.
  }
#ifdef WIN32
  //
  // Initialize winsock.
  //
  wVersionRequested = MAKEWORD(2, 0);
  if (WSAStartup(wVersionRequested, &wsaData) != 0) {
    printf("Could not init WinSock.\n");
    return 1;
  }
#endif
  //
  // Create datagram socket for receiving answers
  // from target and packets sent by target do not
  // cause destination unreachable ICMPs sent back.
  //
  hSock = socket(AF_INET, SOCK_DGRAM, 0);
  if (hSock == -1) {
    printf ("Could not create socket.");
    r = 1;  // Error.
    goto CleanUp;
  }
  //
  // Set the receive buffer size for better performance and stability.
  // With a too small receive buffer data might be dropped.
  //
  if (setsockopt(hSock, SOL_SOCKET, SO_RCVBUF, (char*)&BufSize, sizeof(BufSize)) != 0) {
    printf ("Unable to set receive buffer size.");
    r = 1;  // Error.
    goto CleanUp;
  }
  //
  // Bind to specific port for incoming packets.
  //
  memset(&Addr, 0, sizeof(Addr));
  Addr.sin_family      = AF_INET;
  Addr.sin_port        = htons(PORT);
  Addr.sin_addr.s_addr = INADDR_ANY;
  bind(hSock, (struct sockaddr *)&Addr, sizeof(Addr));
  //
  // Prepare addr. to send to.
  //
  memset(&Addr, 0, sizeof(Addr));
  Addr.sin_family      = AF_INET;
  Addr.sin_port        = htons(PORT);
  Addr.sin_addr.s_addr = htonl(DestIP);
  while (1) {
    PacketCnt = 0;
    printf("\nStarting test (target sends), run time: %d seconds\n", TEST_DURATION);
    //
    // Send client handle.
    //
    aData[0] = 'H';
    r = sendto(hSock, &aData[0], sizeof(aData), 0, (struct sockaddr*)&Addr, sizeof(Addr));
    if (r == -1) {
      printf("sendto() error\n");
      r = 1;  // Error.
      goto CleanUp;
    }
    //
    // Wait for server handle.
    //
    AddrLen = sizeof(Addr);
    r = recvfrom(hSock, &aData[0], 1, 0, (struct sockaddr*)&Addr, &AddrLen);
    if (r != 1) {
      printf("Error while waiting for server handle\n");
      r = 1;  // Error.
      goto CleanUp;
    }
    if (aData[0] != ('H' + 1)) {
      printf("Wrong server handle received\n");
      r = 1;  // Error.
      goto CleanUp;
    }
    //
    // Tell target to send for (TEST_DURATION * 1000) ms.
    //
    aData[0] = 'H';                       // Client handle.
    aData[1] = 'S';                       // Server shall send data.
    memcpy(&aData[2], &TestDuration, 4);  // Tell the server how long to send data.
    sendto(hSock, &aData[0], 6, 0, (struct sockaddr*)&Addr, sizeof(Addr));
    //
    // Fetch data sent to us and wait for final packet.
    //
    while (1) {
      r = recvfrom(hSock, &_abBuffer[0], sizeof(_abBuffer), 0, NULL, NULL);
      if (r <= 0) {
        r = 1;  // Error.
        goto CleanUp;
      }
      if (_abBuffer[1] == ('S' + 2)) {  // Last packet with results ?
        memcpy(&PacketSize    , &_abBuffer[2], 4);
        memcpy(&NumPacketsSent, &_abBuffer[6], 4);
        printf("Packet size used (incl. headers)      : %d bytes\n", PacketSize);
        printf("Data transferred (incl. headers)      : %d bytes\n", (PacketSize * NumPacketsSent));
        printf("Average transfer speed                : %d bytes/s\n", (PacketSize * NumPacketsSent) / TEST_DURATION);
        printf("Packets server sent/client received   : %d/%d packets\n", NumPacketsSent, PacketCnt);
        printf("Average receive speed                 : %d bytes/s\n", (PacketCnt * PacketSize) / TEST_DURATION);
        printf("Packet loss                           : %d Packets (%.4f%%)\n", (NumPacketsSent - PacketCnt), ((float)(((float)NumPacketsSent) / ((float)PacketCnt)) - 1));
        break;
      } else {                          // Packet with test data.
        PacketCnt++;
      }
    }
    printf("\nPress any key to repeat the test, 'x' to exit.\n");
    Key = getch();
    if (Key == 'x') {
      break;
    }
  }
CleanUp:
  if (hSock) {
    CLOSESOCKET(hSock);
  }
#ifdef WIN32
  WSACleanup();
#endif
  return 0;
}

/*************************** end of file ****************************/
